/*
 * Copyright 2017-present Open Networking Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.onosproject.fpcagent.rest;

import com.fasterxml.jackson.databind.JsonNode;
import org.apache.commons.lang.exception.ExceptionUtils;
import org.apache.http.HttpResponse;
import org.apache.http.impl.nio.client.CloseableHttpAsyncClient;
import org.apache.http.impl.nio.client.HttpAsyncClients;
import org.apache.http.nio.IOControl;
import org.apache.http.nio.client.methods.AsyncCharConsumer;
import org.apache.http.nio.client.methods.HttpAsyncMethods;
import org.apache.http.nio.protocol.HttpAsyncRequestProducer;
import org.apache.http.protocol.HttpContext;
import org.onosproject.fpcagent.util.eventStream.SseManager;
import org.onosproject.fpcagent.util.eventStream.ParseStream;
import org.onosproject.rest.AbstractWebResource;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.ws.rs.Consumes;
import javax.ws.rs.POST;
import javax.ws.rs.Path;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import java.io.InputStream;
import java.nio.CharBuffer;
import java.util.AbstractMap;
import java.util.Map;
import java.util.concurrent.Executors;

@Path("/")
public class FpcWebResource extends AbstractWebResource {
    private static final Logger log = LoggerFactory.getLogger(FpcWebResource.class);

    @POST
    @Consumes(MediaType.APPLICATION_JSON)
    @Path("/response")
    public Response getResponse(InputStream stream) {
        JsonNode jsonNode = mapper().valueToTree(stream);
        JsonNode clientUri = jsonNode.get("client-uri");

        log.info("Received {}", clientUri);
        if (!clientUri.isMissingNode()) {
            String uri = clientUri.asText();
            Map<String, CloseableHttpAsyncClient> clients = SseManager.getInstance().clientUriToHttpClient;
            if (!clients.containsKey(uri)) {
                Executors.newSingleThreadExecutor().submit(() -> {
                    try {
                        CloseableHttpAsyncClient client = HttpAsyncClients.createDefault();
                        clients.put(uri, client);
                        HttpAsyncRequestProducer get = HttpAsyncMethods.createGet(uri);
                        client.execute(get, new MyResponseConsumer(uri), null);
                    } catch (Exception e) {
                        log.error(ExceptionUtils.getFullStackTrace(e));
                    }
                });
            }
        }

        return Response.ok()
                .header("Content-Type", "text/event-stream")
                .header("Cache-Control", "no-cache, no-store")
                .header("Connection", "keep-alive")
                .build();
    }

    @POST
    @Consumes(MediaType.APPLICATION_JSON)
    @Path("/request")
    public Response getRequest(InputStream stream) {
        JsonNode jsonNode = mapper().valueToTree(stream);

        return Response.ok()
                .header("Content-Type", "text/event-stream")
                .header("Cache-Control", "no-cache, no-store")
                .header("Connection", "keep-alive")
                .build();
    }

    @POST
    @Consumes(MediaType.APPLICATION_JSON)
    @Path("/notification")
    public Response getNotification(InputStream stream) {
        JsonNode jsonNode = mapper().valueToTree(stream);
        JsonNode clientId = jsonNode.get("client-id");

        log.info("Received {}", clientId);
        if (!clientId.isMissingNode()) {
            String uri = clientId.asText();
            Map<String, CloseableHttpAsyncClient> clients = SseManager.getInstance().clientUriToHttpClient;
            if (!clients.containsKey(uri)) {
                Executors.newSingleThreadExecutor().submit(() -> {
                    try {
                        CloseableHttpAsyncClient client = HttpAsyncClients.createDefault();
                        clients.put(uri, client);
                        HttpAsyncRequestProducer get = HttpAsyncMethods.createGet(uri);
                        client.execute(get, new MyResponseConsumer(uri), null);
                    } catch (Exception e) {
                        log.error(ExceptionUtils.getFullStackTrace(e));
                    }
                });
            }
        }

        return Response.ok()
                .header("Content-Type", "text/event-stream")
                .header("Cache-Control", "no-cache, no-store")
                .header("Connection", "keep-alive")
                .build();
    }

    /**
     * A character consumer to read incoming characters on the request stream
     */
    static class MyResponseConsumer extends AsyncCharConsumer<Boolean> {
        private String clientUri;

        /**
         * Constructor
         *
         * @param clientUri - URI of the FPC Client
         */
        MyResponseConsumer(String clientUri) {
            this.clientUri = clientUri;
        }

        @Override
        protected void onResponseReceived(final HttpResponse response) {
        }

        @Override
        protected void onCharReceived(final CharBuffer buf, final IOControl ioctrl) {
            char[] charArray = new char[buf.remaining()];
            System.arraycopy(buf.array(), 0, charArray, 0, buf.remaining());
            CharBuffer charBuffer = CharBuffer.wrap(charArray);
            ParseStream.getInstance().send(new AbstractMap.SimpleEntry<>(clientUri, charBuffer));
        }

        @Override
        protected void releaseResources() {
        }

        @Override
        protected Boolean buildResult(final HttpContext context) {
            return Boolean.TRUE;
        }

    }
}

